#!/usr/bin/env perl
use warnings;
use strict;
use diagnostics;

# -------------------------------------------------------------
# Global Variables
# -------------------------------------------------------------

# Change the location of PIN to reflect your installation
my $PIN = "/home/dimitrov/Spring2009/pin-2.5-24110-gcc.4.0.0-ia32_intel64-linux/pin";

my $tool = "MyTokensRegCombineDel";
my $input_file = "accmon_miss.txt";
my $run_command;

my $run_options;
my $size = 65536;
my $enable_read = 0;
my $enable_log = 0;
my $instrumentation_limit = 50000000;
my $num_divide = 1;
my $num_select = 0;
#my $split_threshold = 90;
my $split_threshold = 70;
my $num_repl_split = 20; 

my $num_training_set = 17;
my @training_set = (
					"-- my_bc inputs/Test/array.b",
					"-- my_bc inputs/Test/arrayp.b",
					"-- my_bc inputs/Test/aryprm.b",
					"-- my_bc -l inputs/Test/atan.b",
					"-- my_bc -l inputs/Test/checklib.b",
					"-- my_bc inputs/Test/my_div.b",
					"-- my_bc -l inputs/Test/exp.b",
					"-- my_bc inputs/Test/fact.b",
					"-- my_bc -l inputs/Test/jn.b",
					"-- my_bc -l inputs/Test/ln.b",
					"-- my_bc inputs/Test/mul.b",
					"-- my_bc inputs/Test/raise.b",
					"-- my_bc inputs/Test/signum",
					"-- my_bc -l inputs/Test/sine.b",
					"-- my_bc inputs/Test/sqrt1.b",
					"-- my_bc inputs/Test/sqrt2.b",
					"-- my_bc inputs/Test/sqrt.b"
				   );

my @fail_set = ("-- my_bc script/bad.b");

# -------------------------------------------------------------
# Function prototypes 
# -------------------------------------------------------------
sub read_num_repl();
sub should_split($$$);
sub get_num_regions();
sub train_and_trigger();
sub cleanup();


# -------------------------------------------------------------
# Main program 
# -------------------------------------------------------------

# Make sure the PIN tools are up to date
#system "cp -u \$PIN/WORK4/$tool .";


my $should_split = 1;
my $prev_num_repl = 1000000000; # very large
my $curr_num_repl = 0;

#get_num_regions();
train_and_trigger();
cleanup();

# -------------------------------------------------------------
# end of Main program 
# -------------------------------------------------------------



# -------------------------------------------------------------
# Function definitions 
# -------------------------------------------------------------

sub train_and_trigger(){

	print "*" x 50 . "\n";
	print "Train and trigger bug for each code region\n";
	print "*" x 50 . "\n";

	# Now train the predictors and trigger the bug, for each code region
	for($num_select=0; $num_select<$num_divide; $num_select++){
	
		# initial run, do not load prediction table from file
		$enable_read = 0; 
		$run_options = "$PIN -t $tool -ENABLE_ACCMON_READ $enable_read ".
			   "-ENABLE_ACCMON_LOG $enable_log -ACCMON_DISABLE_LEARNING 0 -NO_SHARED_LIBS 0 ".
			   "-ACCMON_TABLE_SIZE $size -INSTRUMENTATION_LIMIT $instrumentation_limit ".
			   "-ACCMON_NUM_DIVIDE $num_divide -ACCMON_NUM_SELECT $num_select ".
			   "-ENABLE_STACK_RELATIVE 1 -ENABLE_ACCMON 1 -ENABLE_SEP 0 ";
		$run_command = $run_options . " " . $training_set[0];
		print $run_command . "\n";
		system $run_command;

		# training runs, load prediction table from file
		$enable_read = 1;
		for(my $i=1; $i<$num_training_set; $i++){

			$run_options = "$PIN -t $tool -ENABLE_ACCMON_READ $enable_read ".
			   "-ENABLE_ACCMON_LOG $enable_log -ACCMON_DISABLE_LEARNING 0 -NO_SHARED_LIBS 0 ".
			   "-ACCMON_TABLE_SIZE $size -INSTRUMENTATION_LIMIT $instrumentation_limit ".
			   "-ACCMON_NUM_DIVIDE $num_divide -ACCMON_NUM_SELECT $num_select ".
			   "-ENABLE_STACK_RELATIVE 1 -ENABLE_ACCMON 1 -ENABLE_SEP 0 ";
			$run_command = $run_options . " " . $training_set[$i];
			print $run_command . "\n";
			system $run_command;
		}# end for

		# save trained tables
		$run_command = "cp accmon_dump.txt trained.$num_select.accmon_dump.txt";
		print $run_command . "\n";
		system $run_command;

		# triger the bug
		$enable_log = 1;
		$run_options = "$PIN -t $tool -ENABLE_ACCMON_READ $enable_read ".
		   "-ENABLE_ACCMON_LOG $enable_log -TOKEN_TABLE_SIZE 4096 ".
		   "-ACCMON_TABLE_SIZE $size -ENABLE_TOKEN 1 -ACCMON_DISABLE_LEARNING 1 ".
		   "-VIOLATION_NEWCODE 1 -NO_SHARED_LIBS 0 ".
		   "-ACCMON_NUM_DIVIDE $num_divide -ACCMON_NUM_SELECT $num_select ".
		   "-ENABLE_STACK_RELATIVE 1 -ENABLE_ACCMON 1 -ENABLE_SEP 0 ";
		$run_command = $run_options . " " . $fail_set[0];
		print $run_command . "\n";
		system $run_command;

		$run_command = "mv accmon.txt buggy.$num_select.accmon.txt";
		print $run_command . "\n";
		system $run_command;
		$run_command = "mv CrashToken.txt CrashToken.$num_select.accmon.txt";
		print $run_command . "\n";
		system $run_command;
		system 'rm -f accmon_dump.txt';

	}# end for

}# end train_and_trigger

sub get_num_regions(){
	
	# This loop runs the first training program repeatedly 
	# to determine the optimum number of code regions to split
	# the program into. 

	print "*" x 50 . "\n";
	print "Determine number of code regions\n";
	print "*" x 50 . "\n";

	while($should_split == 1){
	
		# set up the execution options
		$run_options = "$PIN -t $tool -ENABLE_ACCMON_READ $enable_read ".
			   "-ENABLE_ACCMON_LOG $enable_log -NO_SHARED_LIBS 0 ".
			   "-ACCMON_TABLE_SIZE $size -INSTRUMENTATION_LIMIT $instrumentation_limit ".
			   "-ACCMON_NUM_DIVIDE $num_divide -ACCMON_NUM_SELECT $num_select ".
		       "-ENABLE_STACK_RELATIVE 1 -ENABLE_ACCMON 1 -ENABLE_SEP 0 ";
		$run_command = $run_options . " " . $training_set[0];

		# execute
		print $run_command . "\n";
		system $run_command;

		# determine if we should split the code into code regions
		$curr_num_repl = read_num_repl();
		print "prev: $prev_num_repl, curr: $curr_num_repl\n";
		$should_split = should_split($prev_num_repl, $curr_num_repl, $num_divide);
		$prev_num_repl = $curr_num_repl;

		# if we are splitting again, double the number of code regions
		if($should_split == 1){
			$num_divide = $num_divide * 2;
		}
	}# end while
}# end get_num_regions

sub cleanup(){
	
	# combine the violations from each code region
	$run_command = "cat buggy.* > accmon.$num_divide.txt";
	print $run_command . "\n";
	system $run_command;

	$run_command = "count_unique_violations.pl accmon.$num_divide.txt accmon.$num_divide.unique.txt";
	print $run_command . "\n";
	system $run_command;

#	$run_command = "rm -f buggy.* accmon.$num_divide.txt sorted_accmon.txt";
#	print $run_command . "\n";
#	system $run_command;

	$run_command = "cat CrashToken.* > accmon.$num_divide.CrashToken.txt";
	print $run_command . "\n";
	system $run_command;
	
#	$run_command = "rm -f CrashToken.*";
#	print $run_command . "\n";
#	system $run_command;
}

sub read_num_repl(){
	my $num_repl;
	open(INFILE, "$input_file") or die "Can't open $input_file: $!";

	# read the number of replacements from the input file input file
	while(<INFILE>){
		if( /num_repl:\s([1234567890]+)/){
			$num_repl = $1;
			last; # stop looping
		}
	}
	close(INFILE);
	return $num_repl;
}# end read_num_repl

# If the number of current replacements is 0, then do not split
# If the number of current replacements is less than 90 percent
# of the previous replacements (we have improvement > 10% ) then
# continue splitting
sub should_split($$$){
	my $prev_repl = shift;
	my $curr_repl = shift;
	my $curr_num_divide = shift;

	if( $curr_repl < $num_repl_split){
		return 0;
	}
	else{
		if( ($curr_repl*100)/$prev_repl < $split_threshold ){
			return 1;
		}
		else{
			return 0;
		}
	}
}# end should_split
